from dataclasses import dataclass
from typing import Any

from ..bound import Bound
from ..errors import ValidationError
from .observe_from_latest_event import ObserveFromLatestEvent


@dataclass
class ObserveEventsOptions:
    recursive: bool
    lower_bound: Bound | None = None
    from_latest_event: ObserveFromLatestEvent | None = None

    def validate(self) -> None:
        if self.lower_bound is not None and not isinstance(self.lower_bound, Bound):
            raise ValidationError(
                'ObserveEventsOptions are invalid: lower_bound must be a Bound object.'
            )

        if self.from_latest_event is not None:
            if self.lower_bound is not None:
                raise ValidationError(
                    'ReadEventsOptions are invalid: '
                    'lowerBoundId and fromLatestEvent are mutually exclusive'
                )

    def to_json(self) -> dict[str, Any]:
        result: dict[str, Any] = {
            'recursive': self.recursive,
        }

        if self.lower_bound is not None:
            result['lowerBound'] = self.lower_bound.to_json()

        if self.from_latest_event is not None:
            result['fromLatestEvent'] = self.from_latest_event.to_json()

        return result



def get_get_headers(api_token: str, with_authorization: bool) -> dict[str, str]:
    headers = {}
    if with_authorization:
        headers['Authorization'] = f'Bearer {api_token}'
    return headers



from dataclasses import dataclass
from enum import Enum


class BoundType(str, Enum):
    INCLUSIVE = "inclusive"
    EXCLUSIVE = "exclusive"


@dataclass
class Bound:
    id: str
    type: BoundType

    def to_json(self) -> dict[str, str]:
        return {
            'id': self.id,
            'type': self.type.value
        }



from .bound import Bound, BoundType
from .client import Client
from .container import Container
from .event import Event, EventCandidate
from .errors import ClientError, CustomError, InternalError, ServerError, ValidationError
from .observe_events import (
    ObserveEventsOptions,
    ObserveFromLatestEvent,
    IfEventIsMissingDuringObserve
)
from .read_events import ReadEventsOptions, ReadFromLatestEvent, IfEventIsMissingDuringRead, Order
from .read_event_types import EventType
from .write_events import Precondition, IsSubjectOnEventId, IsSubjectPristine


__all__ = [
    'Bound', 'BoundType',
    'Client',
    'Container',
    'Event', 'EventCandidate',
    'EventType',
    'ObserveEventsOptions', 'ObserveFromLatestEvent', 'IfEventIsMissingDuringObserve',
    'Precondition', 'IsSubjectOnEventId', 'IsSubjectPristine',
    'ReadEventsOptions', 'ReadFromLatestEvent', 'IfEventIsMissingDuringRead', 'Order',
    'ClientError', 'CustomError', 'InternalError', 'ServerError', 'ValidationError',
]



import logging
import time
from http import HTTPStatus

import docker
import requests
from docker import DockerClient, errors

from .client import Client


class Container:
    def __init__(
        self,
    ):
        self._image_name: str = 'thenativeweb/eventsourcingdb'
        self._image_tag: str = 'latest'
        self._api_token: str = 'secret'
        self._internal_port: int = 3000
        self._container = None
        self._docker_client: DockerClient = docker.from_env()
        self._mapped_port: int | None = None
        self._host = 'localhost'

    def _cleanup_existing_containers(self) -> None:
        try:
            containers = self._docker_client.containers.list(
                filters={'ancestor': f'{self._image_name}:{self._image_tag}'})
        except errors.APIError as e:
            logging.warning('Warning: Error listing existing containers: %s', e)
            return

        for container in containers:
            try:
                container.stop()
            except errors.APIError as e:
                logging.warning('Warning: Error stopping container: %s', e)

            try:
                container.remove()
            except errors.APIError as e:
                logging.warning('Warning: Error removing container: %s', e)

    def _create_container(self) -> None:
        port_bindings = {f'{self._internal_port}/tcp': None}
        self._container = self._docker_client.containers.run(
            f'{self._image_name}:{self._image_tag}',
            command=[
                'run',
                '--api-token',
                self._api_token,
                '--data-directory-temporary',
                '--http-enabled',
                '--https-enabled=false',
            ],
            ports=port_bindings,  # type: ignore
            detach=True,
        )  # type: ignore

    def _extract_port_from_container_info(self, container_info):
        port = None
        valid_mapping = True
        port_mappings = None
        host_port_key = 'HostPort'

        try:
            port_mappings = container_info['NetworkSettings']['Ports'].get(
                f'{self._internal_port}/tcp')
        except KeyError:
            valid_mapping = False

        if valid_mapping and port_mappings and isinstance(port_mappings, list) and port_mappings:
            first_mapping = port_mappings[0]

            if host_port_key in first_mapping:
                port_value = first_mapping[host_port_key]
                port = int(port_value)

        return port

    def _fetch_mapped_port(self) -> None:
        if self._container is None:
            raise RuntimeError('Container failed to start')

        max_retries, retry_delay = 5, 1

        for attempt in range(max_retries):
            if (port := self._try_get_port_from_container()) is not None:
                self._mapped_port = port
                return

            if attempt < max_retries - 1:
                time.sleep(retry_delay)

        self._stop_and_remove_container()
        raise RuntimeError('Failed to determine mapped port')

    def _get_container_info(self):
        if self._container is None:
            return None
        return self._docker_client.api.inspect_container(self._container.id)

    def get_api_token(self) -> str:
        return self._api_token

    def get_base_url(self) -> str:
        if self._container is None:
            raise RuntimeError('Container must be running.')
        return f'http://{self.get_host()}:{self.get_mapped_port()}'

    def get_client(self) -> Client:
        return Client(self.get_base_url(), self.get_api_token())

    def get_host(self) -> str:
        if self._container is None:
            raise RuntimeError('Container must be running.')
        return self._host

    def get_mapped_port(self) -> int:
        if self._container is None or self._mapped_port is None:
            raise RuntimeError('Container must be running.')
        return self._mapped_port

    def is_running(self) -> bool:
        return self._container is not None

    def start(self) -> 'Container':
        if self._container is not None:
            return self

        self._pull_or_get_image()
        self._cleanup_existing_containers()
        self._create_container()
        self._fetch_mapped_port()
        self._wait_for_http('/api/v1/ping', timeout=20)

        return self

    def _pull_or_get_image(self) -> None:
        try:
            self._docker_client.images.pull(self._image_name, self._image_tag)
        except errors.APIError as e:
            self._handle_image_pull_error(e)

    def _handle_image_pull_error(self, error):
        image_name = f"{self._image_name}:{self._image_tag}"
        try:
            self._docker_client.images.get(image_name)
        except errors.ImageNotFound:
            raise RuntimeError(
                f'Could not pull image and no local image available: {error}') from error

        logging.warning("Warning: Could not pull image: %s. Using locally cached image.", error)

    def stop(self) -> None:
        self._stop_and_remove_container()

    def _stop_and_remove_container(self) -> None:
        if self._container is None:
            return

        try:
            self._container.stop()
        except errors.NotFound as e:
            logging.warning("Warning: Container not found while stopping: %s", e)
        except errors.APIError as e:
            logging.warning("Warning: API error while stopping container: %s", e)

        try:
            self._container.remove()
        except errors.NotFound as e:
            logging.warning("Warning: Container not found while removing: %s", e)
        except errors.APIError as e:
            logging.warning("Warning: API error while removing container: %s", e)

        self._container = None
        self._mapped_port = None

    def _try_get_port_from_container(self) -> int | None:
        if not self._container:
            return None

        if not (container_info := self._get_container_info()):
            return None

        return self._extract_port_from_container_info(container_info)

    def _wait_for_http(self, path: str, timeout: int) -> None:
        base_url = self.get_base_url()
        url = f'{base_url}{path}'
        start_time = time.time()

        max_attempts = int(timeout * 2)
        for _ in range(max_attempts):
            if time.time() - start_time >= timeout:
                break

            if self._check_endpoint_available(url):
                return

            time.sleep(0.5)

        self._stop_and_remove_container()
        raise TimeoutError(f'Service failed to become ready within {timeout} seconds')

    def _check_endpoint_available(self, url: str) -> bool:
        try:
            response = requests.get(url, timeout=2)
        except requests.RequestException:
            return False
        return self._check_response_ok(response)

    # pylint: disable=R6301
    def _check_response_ok(self, response) -> bool:
        return response is not None and response.status_code == HTTPStatus.OK

    def with_api_token(self, token: str) -> 'Container':
        self._api_token = token
        return self

    def with_image_tag(self, tag: str) -> 'Container':
        self._image_tag = tag
        return self

    def with_port(self, port: int) -> 'Container':
        self._internal_port = port
        return self



from collections.abc import AsyncGenerator

from types import TracebackType
from typing import Any, TypeVar

from http import HTTPStatus
import json

from .is_heartbeat import is_heartbeat
from .is_stream_error import is_stream_error
from .is_event import is_event
from .parse_raw_message import parse_raw_message
from .read_events import ReadEventsOptions

from .errors import CustomError, InternalError, ServerError, ValidationError
from .event import Event, EventCandidate
from .observe_events import ObserveEventsOptions
from .read_event_types import EventType, is_event_type
from .read_subjects import is_subject

from .write_events import Precondition
from .http_client import HttpClient, Response


T = TypeVar('T')


class Client():
    def __init__(
        self,
        base_url: str,
        api_token: str,
    ):
        self.__http_client = HttpClient(base_url=base_url, api_token=api_token)

    async def __aenter__(self) -> "Client":
        await self.__http_client.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: BaseException | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ) -> None:
        await self.__http_client.__aexit__(exc_type, exc_val, exc_tb)

    async def initialize(self) -> None:
        await self.__http_client.initialize()

    async def close(self) -> None:
        await self.__http_client.close()

    @property
    def http_client(self) -> HttpClient:
        return self.__http_client

    async def ping(self) -> None:
        specversion_field = "specversion"
        type_field = "type"
        ping_received_type = "io.eventsourcingdb.api.ping-received"

        response = await self.http_client.get("/api/v1/ping")
        response_body = bytes.decode(await response.body.read(), encoding="utf-8")

        if response.status_code != HTTPStatus.OK:
            raise ServerError(f"Received unexpected response: {response_body}")

        response_json = json.loads(response_body)
        if (
            isinstance(response_json, dict)
            and specversion_field in response_json
            and type_field in response_json
            and response_json.get(type_field) == ping_received_type
        ):
            return

        raise ServerError(f"Received unexpected response: {response_body}")

    async def verify_api_token(self) -> None:
        request_body = json.dumps({})

        response: Response = await self.http_client.post(
            path='/api/v1/verify-api-token',
            request_body=request_body,
        )
        async with response:
            if response.status_code != HTTPStatus.OK:
                raise ServerError(
                    f'Failed to verify API token: {response}'
                )

            response_data = await response.body.read()
            response_data = bytes.decode(response_data, encoding='utf-8')
            response_json = json.loads(response_data)

            # pylint: disable=R2004
            if not isinstance(response_json, dict) or 'type' not in response_json:
                raise ServerError('Failed to parse response: {response}')

            expected_event_type = 'io.eventsourcingdb.api.api-token-verified'
            if response_json.get('type') != expected_event_type:
                raise ServerError(f'Failed to verify API token: {response}')

    async def write_events(
        self,
        event_candidates: list[EventCandidate],
        preconditions: list[Precondition] = None  # type: ignore
    ) -> list[Event]:
        if preconditions is None:
            preconditions = []

        request_body = json.dumps(
            {
                'events': [event_candidate.to_json() for event_candidate in event_candidates],
                'preconditions': [precondition.to_json() for precondition in preconditions]
            }
        )

        response: Response
        response = await self.http_client.post(
            path='/api/v1/write-events',
            request_body=request_body,
        )

        if response.status_code != HTTPStatus.OK:
            raise ServerError(
                f'Unexpected response status: {response}'
            )

        response_data = await response.body.read()
        response_data = bytes.decode(response_data, encoding='utf-8')
        response_data = json.loads(response_data)

        if not isinstance(response_data, list):
            raise ServerError(
                f'Failed to parse response \'{response_data}\' to list.')

        result = []
        for unparsed_event_context in response_data:
            result.append(Event.parse(unparsed_event_context))
        return result

    async def read_events(
        self,
        subject: str,
        options: ReadEventsOptions
    ) -> AsyncGenerator[Event]:
        request_body = json.dumps({
            'subject': subject,
            'options': options.to_json()
        })
        response: Response = await self.__http_client.post(
            path='/api/v1/read-events',
            request_body=request_body,
        )

        async with response:
            if response.status_code != HTTPStatus.OK:
                raise ServerError(
                    f'Unexpected response status: {response}'
                )
            async for raw_message in response.body:
                message = parse_raw_message(raw_message)

                if is_stream_error(message):
                    raise ServerError(f'{message["payload"]["error"]}.')

                if is_event(message):
                    event = Event.parse(message['payload'])
                    yield event
                    continue

                raise ServerError(
                    f'Failed to read events, an unexpected stream item was received: '
                    f'{message}.'
                )

    async def run_eventql_query(self, query: str) -> AsyncGenerator[Any]:
        request_body = json.dumps({
            'query': query,
        })
        response: Response = await self.__http_client.post(
            path='/api/v1/run-eventql-query',
            request_body=request_body,
        )

        async with response:
            if response.status_code != HTTPStatus.OK:
                raise ServerError(
                    f'Unexpected response status: {response}'
                )
            async for raw_message in response.body:
                message = parse_raw_message(raw_message)

                if is_stream_error(message):
                    raise ServerError(f'{message['payload']['error']}.')
                # pylint: disable=R2004
                if message.get('type') == 'row':
                    payload = message['payload']

                    yield payload
                    continue

                raise ServerError(
                    f'Failed to execute EventQL query, an unexpected stream item was received: '
                    f'{message}.'
                )

    async def observe_events(
        self,
        subject: str,
        options: ObserveEventsOptions
    ) -> AsyncGenerator[Event]:
        request_body = json.dumps({
            'subject': subject,
            'options': options.to_json()
        })

        response: Response = await self.http_client.post(
            path='/api/v1/observe-events',
            request_body=request_body,
        )

        async with response:
            if response.status_code != HTTPStatus.OK:
                raise ServerError(
                    f'Unexpected response status: {response}'
                )
            async for raw_message in response.body:
                message = parse_raw_message(raw_message)

                if is_heartbeat(message):
                    continue

                if is_stream_error(message):
                    raise ServerError(f'{message["payload"]["error"]}.')

                if is_event(message):
                    event = Event.parse(message['payload'])
                    yield event
                    continue

                raise ServerError(
                    f'Failed to read events, an unexpected stream item was received: '
                    f'{message}.'
                )

    async def register_event_schema(self, event_type: str, json_schema: dict) -> None:
        request_body = json.dumps({
            'eventType': event_type,
            'schema': json_schema,
        })

        response: Response = await self.http_client.post(
            path='/api/v1/register-event-schema',
            request_body=request_body,
        )

        async with response:
            if response.status_code != HTTPStatus.OK:
                raise ServerError(
                    f'Unexpected response status: {response} '
                )

    async def read_subjects(
        self,
        base_subject: str
    ) -> AsyncGenerator[str]:
        request_body = json.dumps({
            'baseSubject': base_subject
        })

        response: Response = await self.http_client.post(
            path='/api/v1/read-subjects',
            request_body=request_body,
        )

        async with response:
            if response.status_code != HTTPStatus.OK:
                raise ServerError(
                    f'Unexpected response status: {response}'
                )
            async for raw_message in response.body:
                message = parse_raw_message(raw_message)

                if is_stream_error(message):
                    raise ServerError(message['payload']['error'])

                if is_subject(message):
                    yield message['payload']['subject']
                    continue

                raise ServerError(
                    f'Failed to read subjects, an unexpected stream item '
                    f'was received \'{message}\'.'
                )

    async def read_event_types(self) -> AsyncGenerator[EventType]:
        response: Response
        try:
            response = await self.http_client.post(
                path='/api/v1/read-event-types',
                request_body='',
            )
        except CustomError as custom_error:
            raise custom_error
        except Exception as other_error:
            raise InternalError(str(other_error)) from other_error

        async with response:
            if response.status_code != HTTPStatus.OK:
                raise ServerError(
                    f'Unexpected response status: {response}'
                )
            async for raw_message in response.body:
                message = parse_raw_message(raw_message)

                if is_stream_error(message):
                    raise ServerError(message['payload']['error'])

                if is_event_type(message):
                    event_type: EventType
                    try:
                        event_type = EventType.parse(message['payload'])
                    except ValidationError as validation_error:
                        raise ServerError(str(validation_error)) from validation_error
                    except Exception as other_error:
                        raise InternalError(str(other_error)) from other_error

                    yield event_type
                    continue

                raise ServerError(
                    f'Failed to read event types, an unexpected '
                    f'stream item was received \'{message}\'.'
                )



import json
from enum import Enum
from typing import Any

from .bound import Bound


class EventSourcingDBJSONEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, Bound):
            return o.to_json()
        if isinstance(o, Enum):
            return o.value
        return super().default(o)



from collections.abc import Mapping
from http import HTTPStatus

import aiohttp
from aiohttp import StreamReader

Headers = Mapping[str, str]


class Response:
    def __init__(self, response: aiohttp.ClientResponse):
        self.__response: aiohttp.ClientResponse = response

    async def __aenter__(self):
        # Properly await any async initialization if needed
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if not self.__response.closed:
            self.__response.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.__response.closed:
            self.__response.close()

    @property
    def status_code(self) -> HTTPStatus:
        return HTTPStatus(self.__response.status)

    @property
    def headers(self) -> Headers:
        return self.__response.headers

    @property
    def body(self) -> StreamReader:
        return self.__response.content

    def __str__(self) -> str:
        status_code_text = f'{self.status_code} {HTTPStatus(self.status_code).phrase}'
        header_text = dict(self.headers)
        body_text = self.body.read_nowait().decode("utf-8")
        status_code_text = f'{self.status_code} {HTTPStatus(self.status_code).phrase}'
        return f'status_code={status_code_text}, headers={header_text}, body={body_text}'



from types import TracebackType

import aiohttp
from aiohttp import ClientSession

from ..errors.custom_error import CustomError

from .get_get_headers import get_get_headers
from .get_post_headers import get_post_headers
from .response import Response


class HttpClient:
    def __init__(
        self,
        base_url: str,
        api_token: str,
    ):
        self.__base_url = base_url
        self.__api_token = api_token
        self.__session: ClientSession | None = None

    async def __aenter__(self):
        await self.initialize()
        return self

    async def __aexit__(
        self,
        exc_type: BaseException | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ) -> None:
        await self.close()

    async def initialize(self) -> None:
        self.__session = aiohttp.ClientSession()

    async def close(self):
        if self.__session is not None:
            await self.__session.close()
            self.__session = None

    @staticmethod
    def join_segments(first: str, *rest: str) -> str:
        first_without_trailing_slash = first.rstrip('/')
        rest_joined = '/'.join([segment.strip('/') for segment in rest])

        return f'{first_without_trailing_slash}/{rest_joined}'

    async def post(self, path: str, request_body: str) -> Response:
        if self.__session is None:
            await self.initialize()

        url_path = HttpClient.join_segments(self.__base_url, path)
        headers = get_post_headers(self.__api_token)

        async_response = await self.__session.post(  # type: ignore
            url_path,
            data=request_body,
            headers=headers,
        )

        response = Response(async_response)

        return response

    async def get(
        self,
        path: str,
        with_authorization: bool = True,
    ) -> Response:
        if self.__session is None:
            raise CustomError(
                "HTTP client session not initialized. Call initialize() before making requests.")

        async def __request_executor() -> Response:
            url_path = HttpClient.join_segments(self.__base_url, path)
            headers = get_get_headers(self.__api_token, with_authorization)

            async_response = await self.__session.get(  # type: ignore
                url_path,
                headers=headers,
            )

            response = Response(async_response)

            return response

        return await __request_executor()



from .client_error import ClientError
from .custom_error import CustomError
from .internal_error import InternalError
from .server_error import ServerError
from .validation_error import ValidationError



from dataclasses import dataclass
from typing import TypeVar, Any

from ..errors import ValidationError

Self = TypeVar("Self", bound="EventType")


@dataclass
class EventType:
    event_type: str
    is_phantom: bool
    schema: dict[str, Any] | None = None

    @staticmethod
    def parse(unknown_object: dict) -> "EventType":
        event_type = unknown_object.get('eventType')
        if not isinstance(event_type, str):
            raise ValidationError(
                f"Failed to parse eventType '{event_type}' to str."
            )

        is_phantom = unknown_object.get('isPhantom')
        if not isinstance(is_phantom, bool):
            raise ValidationError(
                f"Failed to parse isPhantom '{is_phantom}' to bool."
            )

        schema = unknown_object.get('schema')
        if schema is not None and not isinstance(schema, (dict)):
            raise ValidationError(
                f"Failed to parse schema '{schema}'. Schema must be dict."
            )

        return EventType(
            event_type=event_type,
            is_phantom=is_phantom,
            schema=schema,
        )

    def __hash__(self):
        # Convert dictionary schema to a hashable form (tuple of items)
        if isinstance(self.schema, dict):
            # Sort items to ensure consistent hashing
            schema_items = tuple(sorted((k, str(v)) for k, v in self.schema.items()))
            return hash((self.event_type, self.is_phantom, schema_items))
        return hash((self.event_type, self.is_phantom, self.schema))



from dataclasses import dataclass
from typing import Any

from ..bound import Bound
from .order import Order
from .read_from_latest_event import ReadFromLatestEvent


@dataclass
class ReadEventsOptions:
    recursive: bool
    order: Order | None = None
    lower_bound: Bound | None = None
    upper_bound: Bound | None = None
    from_latest_event: ReadFromLatestEvent | None = None

    def to_json(self) -> dict[str, Any]:
        json: dict[str, Any] = {
            'recursive': self.recursive
        }

        if self.order is not None:
            json['order'] = self.order.value

        if self.lower_bound is not None:
            json['lowerBound'] = {
                'id': str(self.lower_bound.id),
                'type': self.lower_bound.type
            }

        if self.upper_bound is not None:
            json['upperBound'] = {
                'id': str(self.upper_bound.id),
                'type': self.upper_bound.type
            }

        if self.from_latest_event is not None:
            json['fromLatestEvent'] = self.from_latest_event.to_json()

        return json
